# -*- coding: utf-8 -*-

from __future__ import print_function
from pprint import pprint
from kobo.client import ClientCommand


class Worker_Info(ClientCommand):
    """get worker info"""
    enabled = True
    admin = True


    def options(self):
        self.parser.usage = "%%prog %s worker_name" % self.normalized_name

        self.parser.add_option(
            "--oneline",
            default=False,
            action="store_true",
            help="Display one-line dict output instead of pretty-print"
        )


    def run(self, *args, **kwargs):
        if len(args) != 1:
            self.parser.error("No worker specified")

        username = kwargs.pop("username", None)
        password = kwargs.pop("password", None)
        worker_name = args[0]

        self.set_hub(username, password)
        result = self.hub.client.get_worker_info(worker_name)
        if kwargs.pop("oneline"):
            print(result)
        else:
            pprint(result)



# -*- coding: utf-8 -*-


# Based on http://code.djangoproject.com/wiki/XML-RPC
#
# Credits:
#       Brendan W. McAdams


import sys
import six.moves.xmlrpc_client as xmlrpclib
from six.moves.xmlrpc_server import SimpleXMLRPCDispatcher

from django.conf import settings


__all__ = (
    'DjangoXMLRPCDispatcher',
)


class DjangoXMLRPCDispatcher(SimpleXMLRPCDispatcher):
    def __init__(self, allow_none=True, encoding=None):
        if sys.version_info[:2] == (2, 4):
            # doesn't support extra args in python 2.4
            SimpleXMLRPCDispatcher.__init__(self)
        else:
            SimpleXMLRPCDispatcher.__init__(self, allow_none, encoding)

        self.allow_none = allow_none
        self.encoding = encoding
        self.register_multicall_functions()


    def system_multicall(self, request, call_list):
        for call in call_list:
            # insert request to each param list
            call['params'] = [request] + call['params']

        return SimpleXMLRPCDispatcher.system_multicall(self, call_list)


    def register_module(self, module_name, function_prefix):
        """register all the public functions in a module with prefix prepended"""

        if type(module_name) is str:
            module = __import__(module_name, {}, {}, [""])
        else:
            module = module_name

        if hasattr(module, '__all__'):
            fn_list = module.__all__
        else:
            fn_list = dir(module)

        for fn in fn_list:
            if fn.startswith("_"):
                continue

            function = getattr(module, fn)
            if not callable(function):
                continue

            name = fn
            if function_prefix:
                name = "%s.%s" % (function_prefix, name)
            name = name.replace("__", ".")

            self.register_function(function, name)


    def _marshaled_dispatch(self, request, dispatch_method = None):
        """Dispatches an XML-RPC method from marshalled (XML) data.

        XML-RPC methods are dispatched from the marshalled (XML) data
        using the _dispatch method and the result is returned as
        marshalled data. For backwards compatibility, a dispatch
        function can be provided as an argument (see comment in
        SimpleXMLRPCRequestHandler.do_POST) but overriding the
        existing method through subclassing is the prefered means
        of changing method dispatch behavior.
        """

        data = request.body
        params, method = xmlrpclib.loads(data)

        # add request to params
        params = (request, ) + params

        # generate response
        try:
            if dispatch_method is not None:
                response = dispatch_method(method, params)
            else:
                response = self._dispatch(method, params)
            # wrap response in a singleton tuple
            response = (response,)
            response = xmlrpclib.dumps(response, methodresponse=1, allow_none=self.allow_none, encoding=self.encoding)

        except xmlrpclib.Fault as fault:
            response = xmlrpclib.dumps(fault, allow_none=self.allow_none, encoding=self.encoding)

        except:
            # report exception back to server
            if settings.DEBUG:
                from kobo.tback import Traceback
                response = xmlrpclib.dumps(
                    xmlrpclib.Fault(1, u"%s" % Traceback().get_traceback()),
                    allow_none=self.allow_none, encoding=self.encoding)
            else:
                exc_info = sys.exc_info()[1]
                exc_type = exc_info.__class__.__name__
                response = xmlrpclib.dumps(
                    xmlrpclib.Fault(1, "%s: %s" % (exc_type, exc_info)),
                    allow_none=self.allow_none, encoding=self.encoding)

        return response



# -*- coding: utf-8 -*-

import django

# Only for Django >= 1.7
if 'setup' in dir(django):
    # This has to happen before below imports because they have a hard requirement
    # on settings being loaded before import.
    django.setup()

from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.models import User
from django.core.exceptions import ImproperlyConfigured

try:
    from django.test import override_settings
except:
    # Django <= 1.6
    from django.test.utils import override_settings

from mock import Mock, PropertyMock, patch

from kobo.client.constants import TASK_STATES
from kobo.hub.models import Arch, Channel, Task, Worker

from .utils import DjangoRunner

runner = DjangoRunner()
setup_module = runner.start
teardown_module = runner.stop


class TestAuthView(django.test.TransactionTestCase):

    def setUp(self):
        self.client = django.test.Client()

    def test_login(self):
        response = self.client.get('/auth/login/')
        self.assertEquals(response.status_code, 200)

    def test_krb5login(self):
        response = self.client.get('/auth/krb5login/')
        self.assertEquals(response.status_code, 301)
        self.assertIn(response['Location'], ['http://testserver/home/', '/home/'])

    @override_settings(MIDDLEWARE_CLASSES=[])
    def test_krb5login_missing_middleware(self):
        with self.assertRaises(ImproperlyConfigured):
            self.client.get('/auth/krb5login/')

    def test_krb5login_redirect_to(self):
        response = self.client.get('/auth/krb5login/', {REDIRECT_FIELD_NAME: '/auth/login/'})
        self.assertEquals(response.status_code, 301)
        self.assertIn(response['Location'], ['http://testserver/auth/login/', '/auth/login/'])

    def test_logout(self):
        response = self.client.get('/auth/logout/')
        self.assertEquals(response.status_code, 200)


class TestTaskView(django.test.TransactionTestCase):

    def setUp(self):
        self._fixture_teardown()
        user = User.objects.create(username='testuser')
        arch = Arch.objects.create(name='testarch')
        channel = Channel.objects.create(name='testchannel')
        worker = Worker.objects.create(
            worker_key='mock-worker',
            name='mock-worker',
        )

        worker.arches.add(arch)
        worker.channels.add(channel)
        worker.save()

        self.task1 = Task.objects.create(
            worker=worker,
            arch=arch,
            channel=channel,
            owner=user,
            method='TaskOne',
            state=TASK_STATES['FREE'],
        )

        self.task2 = Task.objects.create(
            worker=worker,
            arch=arch,
            channel=channel,
            owner=user,
            method='TaskTwo',
            state=TASK_STATES['OPEN'],
        )

        self.task3 = Task.objects.create(
            worker=worker,
            arch=arch,
            channel=channel,
            owner=user,
            method='TaskThree',
            state=TASK_STATES['CLOSED'],
        )

        self.task4 = Task.objects.create(
            worker=worker,
            arch=arch,
            channel=channel,
            owner=user,
            method='TaskFour',
            state=TASK_STATES['FAILED'],
        )

        self.client = django.test.Client()

    def test_list(self):
        response = self.client.get('/task/')
        self.assertEquals(response.status_code, 200)
        self.assertTrue(self.task1.method in str(response.content))
        self.assertTrue(self.task2.method in str(response.content))
        self.assertTrue(self.task3.method in str(response.content))
        self.assertTrue(self.task4.method in str(response.content))

    def test_list_running(self):
        response = self.client.get('/task/running/')
        self.assertEquals(response.status_code, 200)
        self.assertTrue(self.task1.method in str(response.content))
        self.assertTrue(self.task2.method in str(response.content))
        self.assertTrue(self.task3.method not in str(response.content))
        self.assertTrue(self.task4.method not in str(response.content))

    def test_list_finished(self):
        response = self.client.get('/task/finished/')
        self.assertEquals(response.status_code, 200)
        self.assertTrue(self.task1.method not in str(response.content))
        self.assertTrue(self.task2.method not in str(response.content))
        self.assertTrue(self.task3.method in str(response.content))
        self.assertTrue(self.task4.method in str(response.content))

    def test_detail(self):
        response = self.client.get('/task/%d/' % self.task1.id)
        self.assertEquals(response.status_code, 200)
        self.assertTrue('#%d: %s' % (self.task1.id, self.task1.method) in str(response.content))


class TestArchView(django.test.TransactionTestCase):

    def setUp(self):
        self._fixture_teardown()
        self.arch1 = Arch.objects.create(name='arch-1', pretty_name='arch-1')
        self.arch2 = Arch.objects.create(name='arch-2', pretty_name='arch-2')

        self.client = django.test.Client()

    def test_list(self):
        response = self.client.get('/info/arch/')
        self.assertEquals(response.status_code, 200)
        self.assertTrue(self.arch1.name in str(response.content))
        self.assertTrue(self.arch2.name in str(response.content))

    def test_detail(self):
        response = self.client.get('/info/arch/%d/' % self.arch1.id)
        self.assertEquals(response.status_code, 200)
        self.assertTrue('#%d: %s' % (self.arch1.id, self.arch1.name) in str(response.content))


class TestChannelView(django.test.TransactionTestCase):

    def setUp(self):
        self._fixture_teardown()
        self.channel1 = Channel.objects.create(name='channel-1')
        self.channel2 = Channel.objects.create(name='channel-2')

        self.client = django.test.Client()

    def test_list(self):
        response = self.client.get('/info/channel/')
        self.assertEquals(response.status_code, 200)
        self.assertTrue(self.channel1.name in str(response.content))
        self.assertTrue(self.channel2.name in str(response.content))

    def test_detail(self):
        response = self.client.get('/info/channel/%d/' % self.channel1.id)
        self.assertEquals(response.status_code, 200)
        self.assertTrue('#%d: %s' % (self.channel1.id, self.channel1.name) in str(response.content))


class TestUserView(django.test.TransactionTestCase):

    def setUp(self):
        self._fixture_teardown()
        self.user1 = User.objects.create(username='user-1')
        self.user2 = User.objects.create(username='user-2')

        self.client = django.test.Client()

    def test_list(self):
        response = self.client.get('/info/user/')
        self.assertEquals(response.status_code, 200)
        self.assertTrue(self.user1.username in str(response.content))
        self.assertTrue(self.user2.username in str(response.content))

    def test_detail(self):
        response = self.client.get('/info/user/%d/' % self.user1.id)
        self.assertEquals(response.status_code, 200)
        self.assertTrue('#%d: %s' % (self.user1.id, self.user1.username) in str(response.content))


class TestWorkerView(django.test.TransactionTestCase):

    def setUp(self):
        self._fixture_teardown()
        self.worker1 = Worker.objects.create(
            worker_key='worker-1',
            name='worker-1',
        )

        self.worker2 = Worker.objects.create(
            worker_key='worker-2',
            name='worker-2',
        )

        self.client = django.test.Client()

    def test_list(self):
        response = self.client.get('/info/worker/')
        self.assertEquals(response.status_code, 200)
        self.assertTrue(self.worker1.name in str(response.content))
        self.assertTrue(self.worker2.name in str(response.content))

    def test_detail(self):
        response = self.client.get('/info/worker/%d/' % self.worker1.id)
        self.assertEquals(response.status_code, 200)
        self.assertTrue('#%d: %s' % (self.worker1.id, self.worker1.name) in str(response.content))



# -*- coding: utf-8 -*-


from kobo.django.menu import menu


__all__ = (
    "MenuMiddleware",
)


class LazyMenu(object):
    """
    @summary: Cached menu object
    """
    def __get__(self, request, obj_type=None):
        if not hasattr(request, "_cached_menu"):
            request._cached_menu = menu.setup(request)
        return request._cached_menu


class MenuMiddleware(object):
    """
    @summary: Middleware for menu object.
    """
    def process_request(self, request):
        """
        @summary: Adds menu to request object
        @param request: http request object
        @type request: django.http.HttpRequest
        """
        request.__class__.menu = LazyMenu()



import os

from django.test.utils import get_runner
from django.conf import settings

# Run test with KOBO_MEMORY_PROFILER=1 to generate memory usage reports from
# tests annotated with @profile.
#
# The point of the memory profiler with tests is to prove that the function
# is memory efficient. When using the profiler, you'll want to verify that
# the peak memory usage shows no significant increase in the annotated test.
if os.environ.get('KOBO_MEMORY_PROFILER', '0') == '1':
    from memory_profiler import profile
else:
    # If memory_profiler is disabled, this is a no-op decorator
    def profile(fn):
        return fn


class DjangoRunner(object):
    """Use this for tests which need an active Django environment
    and database.  Create an instance and start/stop around the
    relevant test(s).

    Ideally, this could be set up as a pytest fixture in conftest.py.
    That doesn't work currently due to https://github.com/pytest-dev/pytest/issues/517 ;
    it clashes with django.testcase.TestCase.setUpClass.
    It can instead be used via setup_module/teardown_module.
    """
    def __init__(self):
        self.runner = None
        self.old_config = None

    def start(self):
        runner_class = get_runner(settings)
        self.runner = runner_class()
        self.runner.setup_test_environment()
        self.old_config = self.runner.setup_databases()

    def stop(self):
        self.runner.teardown_databases(self.old_config)
        self.runner.teardown_test_environment()
        self.runner = None
        self.old_config = None


def data_path(basename):
    """Returns path to a file under 'data' dir."""
    this_dir = os.path.dirname(__file__)
    return os.path.join(this_dir, 'data', basename)


class ArgumentIsInstanceOf(object):

    def __init__(self, classinfo):
        self.classinfo = classinfo

    def __eq__(self, other):
        return isinstance(other, self.classinfo)



# -*- coding: utf-8 -*-


"""
Python-like syntax config parser.

USAGE:
    settings = PyConfigParser()
    settings.load_from_{conf, dict, file, string}(...)
    print settings[var]

CONFIG FILE SYNTAX:
PyConfigParser accepts following python-like syntax:
 - variable = <str, int, float, dict, list, tuple>
 - variable = <other_variable>
 - formatting strings can be used:
   - variable = "%s %s" % (var1, var2)
   - variable = "%(key1)s %(key2)s" % <dict>
 - imports are supported:
   - from <file_without_suffix> import *
   - from <file_without_suffix> import var1, var2
"""


from __future__ import print_function
import os
import fnmatch
import itertools
import keyword
import sys
import token
import tokenize
from six.moves import StringIO

from kobo.exceptions import ImproperlyConfigured


__all__ = (
    "get_dict_value",
    "PyConfigParser",
    "ImproperlyConfigured",
)


def get_dict_value(dictionary, key):
    """Return a value from a dictionary, if not found, use 'glob' keys (*, ? metachars), then use default value with '*' key."""
    if dictionary is None:
        return None

    if type(dictionary) is not dict:
        raise TypeError("Dictionary expected, got %s." % type(dictionary))

    try:
        return dictionary[key]
    except KeyError:
        if isinstance(key, str):
            matches = []
            for pattern in dictionary:
                if pattern == '*' or not isinstance(pattern, str):
                    # exclude '*', because it would match every time
                    continue
                if fnmatch.fnmatchcase(key, pattern):
                    matches.append(pattern)
            if len(matches) == 1:
                return dictionary[matches[0]]
            elif len(matches) > 1:
                raise KeyError("Key matches multiple values: %s" % key)
        if '*' in dictionary:
            return dictionary['*']
        raise


def _type_equal(a, b):
    """Check if two values have the same type and are equal."""
    return type(a) == type(b) and a == b


class PyConfigParser(dict):
    """Python-like syntax config parser."""

    get_dict_value = staticmethod(get_dict_value)

    def __init__(self, config_file_suffix="conf", debug=False):
        self._tok_number = None
        self._tok_value = None
        self._tok_begin = None
        self._tok_end = None
        self._tok_line = None
        self._tok_name = None
        self._config_file_suffix = config_file_suffix
        self._debug = debug
        self._open_file = None
        # list of config files in abspath, includes all imported config files
        self.opened_files = []

    def __getitem__(self, name):
        if name.startswith("_"):
            raise KeyError(name)
        return dict.__getitem__(self, name)

    def __setitem__(self, name, value):
        if name.startswith("_"):
            raise KeyError(name)
        return dict.__setitem__(self, name, value)

    def load_from_file(self, file_name):
        """Load data data from a file."""
        fo = open(file_name, "r")
        data = fo.read()
        fo.close()
        self._open_file = file_name
        self.opened_files.append(os.path.abspath(file_name))
        self.load_from_string(data)

    def load_from_string(self, input_string):
        """Load data from a string."""
        if input_string:
            self._tokens = tokenize.generate_tokens(StringIO(input_string).readline)
            for key, value in self._parse():
                self[key] = value

    def load_from_dict(self, input_dict):
        """Load data from a dictionary."""
        if input_dict is not None:
            self.update(input_dict)

    def load_from_conf(self, conf):
        """Load data from another config."""
        self.load_from_dict(conf)

    def _parse(self):
        """Parse config file and store results to this object."""
        while True:
            self._get_token()

            if self._tok_value == "from":
                self._get_from_import()
                continue

            if keyword.iskeyword(self._tok_value):
                raise SyntaxError("Cannot assign to a python keyword: %s" % self._tok_value)

            if self._tok_name == "ENDMARKER":
                break

            self._assert_token(("NAME", ))
            key = self._tok_value

            self._get_token()
            self._assert_token(("OP", "="))

            value = self._get_value()
            yield key, value

    def _assert_token(self, *args):
        """Check if token has proper name and value.

        *args are tuples (name, value), (name, )
        """
        for i in args:
            if len(i) == 1 and i == (self._tok_name, ):
                return
            if len(i) == 2 and i == (self._tok_name, self._tok_value):
                return
        raise SyntaxError("Invalid syntax: file: %s, begin: %s, end: %s, text: %s" % (self._open_file, self._tok_begin, self._tok_end, self._tok_line))

    def _get_token(self, skip_newline=True):
        """Get a new token from token generator."""
        self._tok_number, self._tok_value, self._tok_begin, self._tok_end, self._tok_line = next(self._tokens)
        self._tok_name = token.tok_name.get(self._tok_number, None)

        if self._debug:
            print("%2s %16s %s" % (self._tok_number, self._tok_name, self._tok_value.strip()))

        # skip some tokens
        if self._tok_name in ["COMMENT", "INDENT", "DEDENT"]:
            self._get_token(skip_newline=skip_newline)

        if skip_newline and self._tok_name in ["NL", "NEWLINE"]:
            self._get_token()

    def _get_NAME(self):
        """Return a NAME token value."""
        if self._tok_value == "False":
            return False

        if self._tok_value == "True":
            return True

        if self._tok_value == "None":
            return None

        # return already defined variable
        try:
            return self[self._tok_value]
        except KeyError:
            raise SyntaxError("Undefined variable %r: file: %s, begin: %s, end: %s, text: %s"
                              % (self._tok_value, self._open_file, self._tok_begin, self._tok_end, self._tok_line))

    def _get_STRING(self):
        """Return a STRING token value."""
        # remove apostrophes or quotation marks
        result = self._tok_value[1:-1]

        # look at next token if "%s" follows the string
        self._tokens, tmp = itertools.tee(self._tokens)
        if next(tmp)[1:2] != ("%", ):
            # just a regular string
            return result

        # string formatting is used
        self._get_token()
        self._assert_token(("OP", "%"))
        values = self._get_value()
        return result % values

    def _get_NUMBER(self, negative=False):
        """Return a NUMER token value."""
        if self._tok_value.find(".") != -1:
            result = float(self._tok_value)
        else:
            result = int(self._tok_value)

        if negative:
            return -result
        return result

    def _get_value(self, get_next=True, basic_types_only=False):
        """Get a value (number, string, other variable value, ...)."""
        if get_next:
            self._get_token()

        self._assert_token(("NAME", ), ("NUMBER", ), ("STRING", ), ("OP", "{"), ("OP", "["), ("OP", "("), ("OP", "-"))

        if (self._tok_name, self._tok_value) == ("OP", "-"):
            self._get_token()
            self._assert_token(("NUMBER", ))
            return self._get_NUMBER(negative=True)

        if self._tok_name in ["NAME", "NUMBER", "STRING"]:
            return getattr(self, "_get_%s" % self._tok_name)()

        if not basic_types_only:
            if (self._tok_name, self._tok_value) == ("OP", "{"):
                return self._get_dict()

            if (self._tok_name, self._tok_value) == ("OP", "["):
                return self._get_list()

            if (self._tok_name, self._tok_value) == ("OP", "("):
                return self._get_tuple()

        self._assert_token(("FOO", ))

    def _get_from_import(self):
        """Parse 'from <config> import <variables/*>' and import <config> data to this object."""

        file_name = ""
        while True:
            self._get_token()
            if (self._tok_name, self._tok_value) == ("NAME", "import"):
                break
            file_name += str(self._tok_value)
        file_name += "." + self._config_file_suffix

        file_name = os.path.join(os.path.dirname(self._open_file), file_name)
        self._assert_token(("NAME", "import"))

        imports = []
        self._get_token()
        while self._tok_name not in ("NL", "NEWLINE"):
            self._assert_token(("NAME", ), ("OP", "*"))
            imports.append(self._tok_value)
            self._get_token(skip_newline=False)
            self._skip_commas(skip_newline=False)

        imported_config = self.__class__(config_file_suffix=self._config_file_suffix, debug=self._debug)
        imported_config.load_from_file(file_name)
        self.opened_files.extend(imported_config.opened_files)

        if "*" in imports:
            self.load_from_dict(imported_config)
        else:
            for key in imports:
                try:
                    self[key] = imported_config[key]
                except KeyError:
                    raise KeyError("Can't import %s from %s." % (key, file_name))

    def _skip_commas(self, skip_newline=True):
        """Skip OP tokens which contain commas."""
        while (self._tok_name, self._tok_value) == ("OP", ","):
            self._get_token(skip_newline)

    def _get_dict(self):
        """Get a dictionary content."""
        result = {}
        while True:
            self._get_token()
            self._skip_commas()

            if (self._tok_name, self._tok_value) == ("OP", "}"):
                break

            key = self._get_value(get_next=False, basic_types_only=True)

            # Check for an already present key. This would silently overwrite
            # the previous value, but most likely this is a user error in the
            # configuration that should be reported.
            if any(_type_equal(key, k) for k in result):
                # The condition can not use `key in result` as that would
                # report a problem with {1: 1, True: True} because True == 1.
                line, _ = self._tok_begin
                raise SyntaxError('Duplicate dict key %r in file %s on line %s'
                                  % (key, self._open_file, line))

            self._get_token()
            self._assert_token(("OP", ":"))

            value = self._get_value()
            result[key] = value

        return result

    def _get_list(self):
        """Get a list content."""
        result = []
        while True:
            self._get_token()
            self._skip_commas()

            if (self._tok_name, self._tok_value) == ("OP", "]"):
                break

            value = self._get_value(get_next=False)
            result.append(value)

        return result

    def _get_tuple(self):
        """Get a tuple content."""
        result = []
        while True:
            self._get_token()
            self._skip_commas()

            if (self._tok_name, self._tok_value) == ("OP", ")"):
                break

            value = self._get_value(get_next=False)
            result.append(value)

        return tuple(result)


# settings is created only if PROJECT_CONFIG_FILE is set
if "PROJECT_CONFIG_FILE" in os.environ:
    try:
        settings = PyConfigParser()
        settings.load_from_file(os.environ.get("PROJECT_CONFIG_FILE"))
    except Exception:
        ex = sys.exc_info()[1]
        raise ImproperlyConfigured("Could not load config file: %s" % ex)



# -*- coding: utf-8 -*-


import six
__all__ = (
    "Enum",
    "EnumItem",
    "State",
    "StateEnum",
    "DictSet",
)


class EnumItem(object):
    """Data wrapper for Enum."""

    def __init__(self, name, help_text=None, **kwargs):
        self.name = name
        self.help_text = help_text or ""
        self.options = kwargs

    def __str__(self):
        return self.name

    def __repr__(self):
        return str(self)

    def __eq__(self, obj):
        return str(self) == str(obj)

    def __getitem__(self, name):
        return self.options[name]

    def get(self, name, default=None):
        return self.options.get(name, default)


class Enum(object):
    """Enumerated list."""

    def __init__(self, *args):
        """
        @param args: items to be enumerated
        @type args: str, EnumItem
        """
        self._items = []
        self._order = []

        # TODO: duplicity!
        for i in args:
            if not issubclass(type(i), EnumItem):
                i = EnumItem(str(i))
            if i in self._order:
                raise ValueError("Duplicite item: %s" % i)
            self._items.append(i)
            self._order.append(str(i))

        self._dict = dict([ (value, i) for i, value in enumerate(self._order) ])

    def __iter__(self):
        return iter(self._order)

    def __getitem__(self, key):
        if type(key) in (int, slice):
            return self._order[key]
        return self._dict[key]

    def get(self, key, default=None):
        try:
            return self[key]
        except (IndexError, KeyError):
            return default

    def get_num(self, key, default=None):
        try:
            return self._dict[key]
        except (IndexError, KeyError):
            return default

    def get_value(self, key, default=None):
        try:
            return self._order[key]
        except (IndexError, KeyError):
            return default

    def get_mapping(self):
        """Return list of (key, value) mappings."""
        return [ (self._dict[key], key) for key in self._order ]

    def get_item(self, key):
        if type(key) is int:
            return self._items[key]
        elif isinstance(key, State):
            return key
        elif isinstance(key, StateEnum):
            return self._current_state
        return self._items[self._dict[key]]

    def get_item_help_text(self, key):
        """Return item's help text."""
        return self.get_item(key).help_text

    def get_item_option(self, key, option, default=None):
        """Return item's option passed as a kwarg in it's constructor."""
        return self.get_item(key).get(option, default)


class State(EnumItem):

    def __init__(self, name, next_states, help_text=None, enter=None, leave=None, check_perms=None, methods=None, **kwargs):
        """
        @param name: state name
        @type name:  str
        @param next_states: list of allowed state transitions
        @type next_states:  [str]
        @param help_text: help text associated to the state
        @type help_text:  str
        @param enter: list of functions to be executed on entering the state, they take (old_state, **kwargs) arguments
        @type enter:  [function]
        @param leave: list of functions to be executed on leaving the state, they take (old_state, **kwargs) arguments
        @type leave:  [function]
        @param methods:
        @type methods:  {"": }
        """

        EnumItem.__init__(self, name, help_text=help_text, **kwargs)
        self.next_states = next_states or []
        self.enter = enter or []
        self.leave = leave or []
        self.check_perms = check_perms or []
        self.methods = methods or {}


    def __getattr__(self, name):
        if name in self.methods:
            return self.methods[name]


class StateEnum(Enum):

    def __init__(self, *args):
        Enum.__init__(self, *args)
        self._current_state = None
        self._to = None

        # check if all states and transitions are sane
        all_next_states = set()
        for state in self._items:
            for next_state in state.next_states:
                 all_next_states.add(next_state)
        for state in all_next_states:
            if state not in self._items:
                 raise ValueError("State '%s' is not defined. Available states: %s" % (state, self._items))

    def __str__(self):
        if self._current_state is None:
            return ""
        return str(self.get_state_id())
#        return self._current_state.name

    def set_state(self, state):
        self._current_state = self.get_item(state)

    def get_state(self):
        return self._current_state

    def get_state_id(self):
        return self.get_num(self._current_state.name)

    def get_final_states(self, return_id_list=False):
        result = []
        for state in self._items:
            if state.next_states:
                continue
            if return_id_list:
                result.append(self.get_num(state.name))
            else:
                result.append(state.name)
        return result

    def get_next_states_mapping(self, append_current=True, user=None, **kwargs):
        states = set()
        current_state = self._current_state.name

        if self._current_state is None:
            return []

        if append_current:
            states.add(current_state)
        states |= set(self._current_state.next_states)

        if user is not None:
            invalid_states = set()
            check_perms_args = dict(transition=True, user=user)
            check_perms_args.update(kwargs)

            for new_state in states:
                if new_state == current_state:
#                    valid_states.append(state)
                    continue

                # check permissions: all must be valid
                for func in self._current_state.check_perms:
                    if not func(current_state, new_state, **check_perms_args):
                        invalid_states.add(new_state)
                        break

            states -= invalid_states

        return sorted([(self.get_num(name), name) for name in states])

    def change_state(self, new_state, commit = True, **kwargs):
        '''new_state is number or state_name
           if commit is set, state is changed,
           otherwise transition is just tested and prepared
           commit with new_state == None just makes prepared transition if
           there is any, otherwise raises exception

           All callbacks are called with 'commit' parameter, to make
           side-effects just once

           returns: True if change was commited, False if not or no change
           was made (same new state, no new state)

           raises:
               ValueError:
                   - for invalid permissions
                   - invalid transitions
                   - commiting not prepared transition
        '''
        if new_state:
            new_state = self.get_item(new_state)
        if new_state == self._current_state or (new_state is None and not commit):
            return False
        if new_state is None and (not commit or self._to is None):
            raise ValueError('No new state set and commit specified')
        if new_state is None:
            new_state = self._to

        current_state = self._current_state.name
        #new_state = self.get_value(new_state)
        if str(new_state) not in self._current_state.next_states:
            raise ValueError("Invalid transition '%s' -> '%s'." % (current_state, new_state))

        # check transition permissions
        # if username is needed check_perms functions, pass it through kwargs
        check_perms_args = dict(commit=commit)
        check_perms_args.update(kwargs)
        for func in self._current_state.check_perms:
            if not func(current_state, new_state, **check_perms_args):
                raise ValueError("Insufficient rights for transition '%s' -> '%s'." % (current_state, new_state))

        # run "leave" functions on current state
        for func in self._current_state.leave:
            func(current_state, new_state, **kwargs)

        # run "enter" functions on new state
        for func in new_state.enter:
            func(current_state, new_state, **kwargs)

        if commit:
            self.set_state(new_state)
            self._to = None
            return True
        else:
            self._to = new_state
        return False


class DictSet(dict):
    """Dictionary with set operations on keys."""

    def copy(self):
        return DictSet(super(DictSet, self).copy())

    def __sub__(self, other):
        result = DictSet()
        for key, value in six.iteritems(self):
            if key not in other:
                result[key] = value
        return result

    def __or__(self, other):
        result = self.copy()
        result.update(other)
        return result

    def __and__(self, other):
        result = DictSet()
        for key, value in six.iteritems(self):
            if key in other:
                result[key] = value
        return result

    def add(self, key, value):
        # return True if a new key was added
        if key in self:
            return False
        self[key] = value
        return True

    def remove(self, key):
        del self[key]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import django

# Only for Django >= 1.7
if 'setup' in dir(django):
    # This has to happen before below imports because they have a hard requirement
    # on settings being loaded before import.
    django.setup()

import unittest
from six import BytesIO

from kobo.hub.models import _tail as tail


SAMPLE_LINES = [
    b'this is a sample',
    b'string;     each',
    b'line    contains',
    b'16 chars exclud-',
    b'ing newline, and',
    b'six lines  total'
]
SAMPLE_STRING = b'\n'.join(SAMPLE_LINES)


class TestTail(unittest.TestCase):

    def test_tail_empty(self):
        """tail of empty object returns empty"""
        (actual, offset) = tail(BytesIO(), 1024, 1024)

        self.assertEqual(actual, b'')
        self.assertEqual(offset, 0)

    def test_tail_noop(self):
        """tail returns all content if it fits in requested size"""
        (actual, offset) = tail(BytesIO(SAMPLE_STRING), 1024, 1024)

        self.assertEqual(actual, SAMPLE_STRING)
        self.assertEqual(offset, len(SAMPLE_STRING))

    def test_tail_limit(self):
        """tail returns trailing lines up to given limit"""
        expected = b'\n'.join([
            b'ing newline, and',
            b'six lines  total',
        ])
        (actual, offset) = tail(BytesIO(SAMPLE_STRING), 40, 1024)

        self.assertEqual(actual, expected)
        self.assertEqual(offset, len(SAMPLE_STRING))

    def test_tail_line_break(self):
        """tail breaks in middle of line if lines are longer than max length"""
        expected = b'\n'.join([
            # this line is partially returned
            b'xclud-',
            b'ing newline, and',
            b'six lines  total',
        ])
        (actual, offset) = tail(BytesIO(SAMPLE_STRING), 40, 10)

        self.assertEqual(actual, expected)
        self.assertEqual(offset, len(SAMPLE_STRING))



import django
from distutils.version import LooseVersion

def django_version_ge(version_str):
    """
    Check if current django is in higher or equal version than specified by
    version_str parameter
    """

    return LooseVersion(django.get_version()) >= LooseVersion(version_str)



#!/usr/bin/python
# -*- coding: utf-8 -*-


import unittest
import six

import logging
from kobo.log import *




class TestLog(unittest.TestCase):
    def setUp(self):
        self.logger = logging.getLogger("TestLogger")

    def test_verbose_hack(self):
        self.logger.verbose("foo")
        logging.verbose("foo")
        self.assertEqual(logging.VERBOSE, 15)
        if six.PY2:
            # There is no _levelNames attribute in Python 3
            self.assertTrue("VERBOSE" in logging._levelNames)
        self.assertEqual(logging.getLevelName(15), "VERBOSE")



from django.conf.urls import url, include
from django.contrib import admin
from django.http import HttpResponse


def home(request):
    return HttpResponse("Index", status=200, content_type="text/plain")


urlpatterns = [
    url(r'^admin/', admin.site.urls),
    url(r"^auth/", include("kobo.hub.urls.auth")),
    url(r"^home/$", home, name="home/index"),
    url(r"^task/", include("kobo.hub.urls.task")),
    url(r"^info/arch/", include("kobo.hub.urls.arch")),
    url(r"^info/channel/", include("kobo.hub.urls.channel")),
    url(r"^info/user/", include("kobo.hub.urls.user")),
    url(r"^info/worker/", include("kobo.hub.urls.worker")),
]



# -*- coding: utf-8 -*-

import unittest2 as unittest

from mock import Mock, PropertyMock, patch, call

from kobo.client.constants import TASK_STATES
from kobo.worker.task import TaskBase, FailTaskException


class ForegroundTaskBase(TaskBase):
    foreground = True


class TestTaskBase(unittest.TestCase):

    def test_taskbase(self):
        task_info = {'id': 100}
        hub = Mock(worker=Mock(get_task=Mock(return_value=task_info)))
        conf = {'key': 'value'}
        task_id = 100
        args = ('value', 100)

        t = TaskBase(hub, conf, task_id, args)
        self.assertTrue(t.enabled)
        self.assertEqual(t.hub, hub)
        self.assertEqual(t.conf, conf)
        self.assertEqual(t.task_id, task_id)
        self.assertEqual(t.task_info, task_info)

        hub.worker.get_task.assert_called_once_with(task_id)

    def test_task_manager_background_task(self):
        task_info = {'id': 100}
        hub = Mock(worker=Mock(get_task=Mock(return_value=task_info)))
        conf = {'key': 'value'}
        task_id = 100
        args = ('value', 100)

        t = TaskBase(hub, conf, task_id, args)

        tm = t.task_manager
        self.assertIsNone(tm)

        with self.assertRaises(ValueError):
            t.task_manager = Mock()

        tm = t.task_manager
        self.assertIsNone(tm)

    def test_task_manager_foreground_task(self):
        task_info = {'id': 100}
        hub = Mock(worker=Mock(get_task=Mock(return_value=task_info)))
        conf = {'key': 'value'}
        task_id = 100
        args = ('value', 100)

        # task_manager functions check `__class__`
        # and it's impossible to safely mock it.
        t = ForegroundTaskBase(hub, conf, task_id, args)

        tm = t.task_manager
        self.assertIsNone(tm)

        t.task_manager = Mock()

        tm = t.task_manager
        self.assertIsNotNone(tm)

    def test_args(self):
        task_info = {'id': 100}
        hub = Mock(worker=Mock(get_task=Mock(return_value=task_info)))
        conf = {'key': 'value'}
        task_id = 100
        args = []

        t = TaskBase(hub, conf, task_id, args)
        self.assertEqual(len(t._args), 0)

        # access directly
        t._args.append(100)
        self.assertEqual(len(t._args), 1)

        # copy to prevent modification
        t.args.append(200)
        self.assertEqual(len(t._args), 1)

    def test_subtask_list(self):
        task_info = {'id': 100}
        hub = Mock(worker=Mock(get_task=Mock(return_value=task_info)))
        conf = {'key': 'value'}
        task_id = 100
        args = []

        t = TaskBase(hub, conf, task_id, args)
        self.assertEqual(len(t._subtask_list), 0)

        # access directly
        t._subtask_list.append({'id': 1})
        self.assertEqual(len(t._subtask_list), 1)

        # copy to prevent modification
        t.subtask_list.append({'id': 2})
        self.assertEqual(len(t._subtask_list), 1)

    def test_run(self):
        task_info = {'id': 100}
        hub = Mock(worker=Mock(get_task=Mock(return_value=task_info)))
        conf = {'key': 'value'}
        task_id = 100
        args = []

        t = TaskBase(hub, conf, task_id, args)

        with self.assertRaises(NotImplementedError):
            t.run()

    def test_fail(self):
        task_info = {'id': 100}
        hub = Mock(worker=Mock(get_task=Mock(return_value=task_info)))
        conf = {'key': 'value'}
        task_id = 100
        args = []

        t = TaskBase(hub, conf, task_id, args)

        with self.assertRaises(FailTaskException):
            t.fail()

    def test_spawn_subtask(self):
        task_info = {'id': 100}
        conf = {'key': 'value'}
        task_id = 100
        subtask_id = 999
        args = []
        hub = Mock(worker=Mock(
            get_task=Mock(return_value=task_info),
            create_subtask=Mock(return_value=subtask_id),
        ))

        t = TaskBase(hub, conf, task_id, args)
        t.foreground = False

        ret_id = t.spawn_subtask('method', [], 'label')
        self.assertEqual(ret_id, subtask_id)

        hub.worker.create_subtask.assert_called_once_with('label', 'method', [], task_id)

    def test_spawn_subtask_foreground_task(self):
        task_info = {'id': 100}
        hub = Mock(worker=Mock(get_task=Mock(return_value=task_info)))
        conf = {'key': 'value'}
        task_id = 100
        args = []

        t = TaskBase(hub, conf, task_id, args)
        t.foreground = True

        with self.assertRaises(RuntimeError):
            t.spawn_subtask('method', [])

        hub.worker.create_subtask.assert_not_called()

    def test_wait_no_subtasks(self):
        task_info = {'id': 100}
        conf = {'key': 'value'}
        task_id = 100
        args = []
        hub = Mock(worker=Mock(
            get_task=Mock(return_value=task_info),
            check_wait=Mock(return_value=([], [])),
        ))

        t = TaskBase(hub, conf, task_id, args)
        t.foreground = False

        finished = t.wait()
        self.assertEqual(len(finished), 0)

        hub.worker.check_wait.assert_called_once_with(task_id)
        hub.worker.get_task.assert_called_once_with(task_id)
        hub.worker.wait.assert_called_once_with(task_id, None)

    def test_wait_with_subtask_list_closed(self):
        def get_task(task_id):
            return {'id': task_id, 'state': TASK_STATES['CLOSED']}

        conf = {'key': 'value'}
        task_id = 100
        subtask_id = 999
        args = []
        hub = Mock(worker=Mock(
            get_task=Mock(side_effect=get_task),
            check_wait=Mock(return_value=([subtask_id], [])),
            create_subtask=Mock(return_value=subtask_id),
        ))

        t = TaskBase(hub, conf, task_id, args)
        t.foreground = False

        t.spawn_subtask('method', [])

        finished = t.wait(subtask_id)
        self.assertEqual(len(finished), 1)

        hub.worker.check_wait.assert_called_once_with(task_id)
        hub.worker.get_task.assert_has_calls([call(task_id), call(subtask_id)], any_order=False)
        hub.worker.wait.assert_called_once_with(task_id, [subtask_id])

    def test_wait_with_subtask_closed(self):
        def get_task(task_id):
            return {'id': task_id, 'state': TASK_STATES['CLOSED']}

        conf = {'key': 'value'}
        task_id = 100
        subtask_id = 999
        args = []
        hub = Mock(worker=Mock(
            get_task=Mock(side_effect=get_task),
            check_wait=Mock(return_value=([subtask_id], [])),
            create_subtask=Mock(return_value=subtask_id),
        ))

        t = TaskBase(hub, conf, task_id, args)
        t.foreground = False

        t.spawn_subtask('method', [])

        finished = t.wait()
        self.assertEqual(len(finished), 1)

        hub.worker.check_wait.assert_called_once_with(task_id)
        hub.worker.get_task.assert_has_calls([call(task_id), call(subtask_id)], any_order=False)
        hub.worker.wait.assert_called_once_with(task_id, None)

    def test_wait_with_subtask_timeout(self):
        def get_task(task_id):
            return {'id': task_id, 'state': TASK_STATES['TIMEOUT']}

        conf = {'key': 'value'}
        task_id = 100
        subtask_id = 999
        args = []
        hub = Mock(worker=Mock(
            get_task=Mock(side_effect=get_task),
            check_wait=Mock(return_value=([subtask_id], [])),
            create_subtask=Mock(return_value=subtask_id),
        ))

        t = TaskBase(hub, conf, task_id, args)
        t.foreground = False

        t.spawn_subtask('method', [])

        with self.assertRaises(FailTaskException):
            t.wait()

        hub.worker.check_wait.assert_called_once_with(task_id)
        hub.worker.get_task.assert_has_calls([call(task_id), call(subtask_id)], any_order=False)
        hub.worker.wait.assert_called_once_with(task_id, None)

    def test_wait_foreground_task(self):
        task_info = {'id': 100}
        hub = Mock(worker=Mock(get_task=Mock(return_value=task_info)))
        conf = {'key': 'value'}
        task_id = 100
        args = []

        t = TaskBase(hub, conf, task_id, args)
        t.foreground = True

        with self.assertRaises(RuntimeError):
            t.wait()

        hub.worker.wait.assert_not_called()



# -*- coding: utf-8 -*-

import tempfile
import signal

import pytest
import six
import unittest2 as unittest

from mock import Mock, patch, call

from kobo.exceptions import ShutdownException
from kobo.worker import main


class DummyTaskManager(object):

    def __init__(self, conf, logger):
        def get_next_task_mock():
            self.run_count += 1

            if self.run_count >= self.max_runs:
                raise ShutdownException()

            if self.fail:
                raise Exception('This task always fails.')

        self.run_count = 0
        self.max_runs = conf.get('max_runs', 1)
        self.fail = conf.get('fail', False)
        self._logger = logger
        self.hub = Mock(spec=['_login'], _login=Mock(return_value=None))
        self.sleep = Mock(return_value=None)
        self.shutdown = Mock(return_value=None)
        self.get_next_task = Mock(side_effect=get_next_task_mock)
        self.update_tasks = Mock(return_value=None)
        self.update_worker_info = Mock(return_value=None)
        self.log_debug = Mock(return_value=None)
        self.log_error = Mock(return_value=None)
        self.log_info = Mock(return_value=None)


class TestMainLoop(unittest.TestCase):

    def setUp(self):
        self.task_manager = None

    def tearDown(self):
        self.task_manager = None

    def create_task_manager(self):
        def wrap(conf, logger):
            self.task_manager = DummyTaskManager(conf=conf, logger=logger)
            return self.task_manager
        return wrap

    def test_daemon_shutdown(self):
        with self.assertRaises(ShutdownException):
            main.daemon_shutdown()

    def test_main_loop_task_manager_exception(self):
        with patch('kobo.worker.main.TaskManager', Mock(side_effect=ValueError)) as mock_tm:
            with patch.object(main.signal, 'signal') as signal_mock:
                with patch.object(main.sys, 'exit') as exit_mock:
                    with self.assertRaises(ValueError):
                        main.main_loop({}, foreground=False)

                        mock_tm.assert_called_once()
                        signal_mock.assert_called_once_with(signal.SIGTERM, main.daemon_shutdown)
                        exit_mock.assert_not_called()

    def test_main_loop_task_exception(self):
        max_runs = 10

        with patch('kobo.worker.main.TaskManager', self.create_task_manager()):
            with patch.object(main.signal, 'signal') as signal_mock:
                with patch.object(main.sys, 'exit') as exit_mock:
                    main.main_loop({
                        'max_runs': max_runs,
                        'fail': True,
                    }, foreground=False)

                    signal_mock.assert_has_calls([
                        call(signal.SIGTERM, main.daemon_shutdown),
                        call(signal.SIGINT, signal.SIG_IGN),
                        call(signal.SIGTERM, signal.SIG_IGN),
                    ], any_order=False)

                    exit_mock.assert_not_called()

            self.assertEqual(self.task_manager.hub._login.call_count, max_runs)
            self.assertEqual(self.task_manager.update_worker_info.call_count, max_runs)
            self.assertEqual(self.task_manager.update_tasks.call_count, max_runs)
            self.assertEqual(self.task_manager.get_next_task.call_count, max_runs)
            self.assertEqual(self.task_manager.sleep.call_count, max_runs - 1)
            self.assertEqual(self.task_manager.log_error.call_count, max_runs - 1)
            self.assertEqual(self.task_manager.shutdown.call_count, 1)

    def test_main_loop(self):
        max_runs = 10

        with patch('kobo.worker.main.TaskManager', self.create_task_manager()):
            with patch.object(main.signal, 'signal') as signal_mock:
                with patch.object(main.sys, 'exit') as exit_mock:
                    main.main_loop({
                        'max_runs': max_runs,
                    }, foreground=False)

                    signal_mock.assert_has_calls([
                        call(signal.SIGTERM, main.daemon_shutdown),
                        call(signal.SIGINT, signal.SIG_IGN),
                        call(signal.SIGTERM, signal.SIG_IGN),
                    ], any_order=False)

                    exit_mock.assert_not_called()

            self.assertEqual(self.task_manager.hub._login.call_count, max_runs)
            self.assertEqual(self.task_manager.update_worker_info.call_count, max_runs)
            self.assertEqual(self.task_manager.update_tasks.call_count, max_runs)
            self.assertEqual(self.task_manager.get_next_task.call_count, max_runs)
            self.assertEqual(self.task_manager.sleep.call_count, max_runs - 1)
            self.assertEqual(self.task_manager.shutdown.call_count, 1)

    def test_main_loop_in_foreground(self):
        max_runs = 10

        with patch('kobo.worker.main.TaskManager', self.create_task_manager()):
            with patch.object(main.signal, 'signal') as signal_mock:
                with patch.object(main.sys, 'exit') as exit_mock:
                    main.main_loop({
                        'max_runs': max_runs,
                    }, foreground=False)

                    signal_mock.assert_has_calls([
                        call(signal.SIGTERM, main.daemon_shutdown),
                        call(signal.SIGINT, signal.SIG_IGN),
                        call(signal.SIGTERM, signal.SIG_IGN),
                    ], any_order=False)

                    exit_mock.assert_not_called()

            self.assertEqual(self.task_manager.hub._login.call_count, max_runs)
            self.assertEqual(self.task_manager.update_worker_info.call_count, max_runs)
            self.assertEqual(self.task_manager.update_tasks.call_count, max_runs)
            self.assertEqual(self.task_manager.get_next_task.call_count, max_runs)
            self.assertEqual(self.task_manager.sleep.call_count, max_runs - 1)
            self.assertEqual(self.task_manager.shutdown.call_count, 1)

    def test_main_loop_with_file_logger(self):
        max_runs = 1

        with patch('kobo.worker.main.TaskManager', self.create_task_manager()):
            with patch.object(main.kobo, 'log') as log_mock:
                with patch.object(main.signal, 'signal') as signal_mock:
                    with patch.object(main.sys, 'exit') as exit_mock:
                        main.main_loop({
                            'max_runs': max_runs,
                            'LOG_FILE': '/tmp/log.txt',
                        }, foreground=False)

                        log_mock.add_rotating_file_logger.assert_called_once_with(
                            self.task_manager._logger,
                            '/tmp/log.txt',
                            log_level=10,
                        )

                        log_mock.add_stderr_logger.assert_not_called()

                        signal_mock.assert_has_calls([
                            call(signal.SIGTERM, main.daemon_shutdown),
                            call(signal.SIGINT, signal.SIG_IGN),
                            call(signal.SIGTERM, signal.SIG_IGN),
                        ], any_order=False)

                        exit_mock.assert_not_called()

            self.assertEqual(self.task_manager.hub._login.call_count, max_runs)
            self.assertEqual(self.task_manager.update_worker_info.call_count, max_runs)
            self.assertEqual(self.task_manager.update_tasks.call_count, max_runs)
            self.assertEqual(self.task_manager.get_next_task.call_count, max_runs)
            self.assertEqual(self.task_manager.sleep.call_count, max_runs - 1)
            self.assertEqual(self.task_manager.shutdown.call_count, 1)

    def test_main_loop_foreground_with_file_logger(self):
        max_runs = 1

        with patch('kobo.worker.main.TaskManager', self.create_task_manager()):
            with patch.object(main.kobo, 'log') as log_mock:
                with patch.object(main.signal, 'signal') as signal_mock:
                    with patch.object(main.sys, 'exit') as exit_mock:
                        main.main_loop({
                            'max_runs': max_runs,
                            'LOG_FILE': '/tmp/log.txt',
                        }, foreground=True)

                        log_mock.add_rotating_file_logger.assert_called_once_with(
                            self.task_manager._logger,
                            '/tmp/log.txt',
                            log_level=10,
                        )

                        log_mock.add_stderr_logger.assert_called_once_with(self.task_manager._logger)

                        signal_mock.assert_has_calls([
                            call(signal.SIGTERM, main.daemon_shutdown),
                            call(signal.SIGINT, signal.SIG_IGN),
                            call(signal.SIGTERM, signal.SIG_IGN),
                        ], any_order=False)

                        exit_mock.assert_not_called()

            self.assertEqual(self.task_manager.hub._login.call_count, max_runs)
            self.assertEqual(self.task_manager.update_worker_info.call_count, max_runs)
            self.assertEqual(self.task_manager.update_tasks.call_count, max_runs)
            self.assertEqual(self.task_manager.get_next_task.call_count, max_runs)
            self.assertEqual(self.task_manager.sleep.call_count, max_runs - 1)
            self.assertEqual(self.task_manager.shutdown.call_count, 1)


class TestMain(unittest.TestCase):

    def test_main_with_no_commands(self):
        conf = {'PID_FILE': '/tmp/pid'}

        with patch.object(main.kobo.process, 'daemonize') as daemonize_mock:
            with patch.object(main.os, 'kill') as kill_mock:
                with patch.object(main.sys, 'exit') as exit_mock:
                    main.main(conf, argv=[])

                    exit_mock.assert_not_called()
                    kill_mock.assert_not_called()
                    daemonize_mock.assert_called_once_with(
                        main.main_loop,
                        conf=conf,
                        daemon_pid_file='/tmp/pid',
                        foreground=False,
                    )

    def test_main_foreground_command(self):
        conf = {'PID_FILE': '/tmp/pid'}

        with patch('kobo.worker.main.main_loop') as main_loop_mock:
            with patch.object(main.kobo.process, 'daemonize') as daemonize_mock:
                with patch.object(main.os, 'kill') as kill_mock:
                    with patch.object(main.sys, 'exit') as exit_mock:
                        main.main(conf, argv=['--foreground'])

                        exit_mock.assert_not_called()
                        kill_mock.assert_not_called()
                        daemonize_mock.assert_not_called()
                        main_loop_mock.assert_called_once_with(conf, foreground=True)

    def test_main_pid_file_command(self):
        with patch.object(main.kobo.process, 'daemonize') as daemonize_mock:
            with patch.object(main.os, 'kill') as kill_mock:
                with patch.object(main.sys, 'exit') as exit_mock:
                    main.main({}, argv=['--pid-file', '/tmp/pid'])

                    exit_mock.assert_not_called()
                    kill_mock.assert_not_called()
                    daemonize_mock.assert_called_once_with(
                        main.main_loop,
                        conf={},
                        daemon_pid_file='/tmp/pid',
                        foreground=False,
                    )

    def test_main_kill_command(self):
        with tempfile.NamedTemporaryFile('r+') as temp_file:
            temp_file.write('999')
            temp_file.flush()

            with patch.object(main.kobo.process, 'daemonize') as daemonize_mock:
                with patch.object(main.os, 'kill') as kill_mock:
                    with patch.object(main.sys, 'exit') as exit_mock:
                        main.main({}, argv=['--pid-file', temp_file.name, '--kill'])

                        exit_mock.assert_called_once_with(0)
                        kill_mock.assert_called_once_with(999, 15)
                        # since exit is mocked and don't exits the program daemonize will be called.
                        daemonize_mock.assert_called_once()



# Settings for Django testcases against kobo hub
import os
import kobo
from django import VERSION

KOBO_DIR = os.path.normpath(
    os.path.join(os.path.dirname(__file__), '..', 'kobo')
)

SECRET_KEY = "key"
XMLRPC_METHODS = []
TASK_DIR = '/tmp/kobo-test-tasks'
UPLOAD_DIR = '/tmp/kobo-test-dir'

# The middleware and apps below are the bare minimum required
# to let kobo.hub load successfully

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'kobo.django.auth.middleware.LimitedRemoteUserMiddleware',
    'kobo.hub.middleware.WorkerMiddleware',
)

INSTALLED_APPS = (
    'django.contrib.contenttypes',
    'django.contrib.staticfiles',
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.sessions',
    'kobo.django',
    'kobo.hub',
)

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'testdatabase',
    }
}

# We need to specify the template dirs because:
# - the admin/templates don't belong to a particular django app, instead
#   they're intended to be copied to another app by a custom command, but
#   we don't want to run that during the tests
# - automatic lookup of templates under kobo/hub isn't working for some reason,
#   not sure why, but might be related to use of deprecated arguments in
#   render_to_string (FIXME)
#
# The way to specify the template dirs differs between newer and older versions of Django
if VERSION[0:3] < (1, 9, 0):
    TEMPLATE_DIRS = (
        os.path.join(KOBO_DIR, 'admin/templates/hub/templates'),
        os.path.join(KOBO_DIR, 'hub/templates'),
    )
if VERSION[0:3] >= (1, 9, 0):
    TEMPLATES = [
        {
            'BACKEND': 'django.template.backends.django.DjangoTemplates',
            'DIRS': (
                os.path.join(KOBO_DIR, 'admin/templates/hub/templates'),
                os.path.join(KOBO_DIR, 'hub/templates')
            ),
            'APP_DIRS': True
        }
    ]

ROOT_URLCONF = 'tests.hub_urls'

STATIC_URL = os.path.join(os.path.dirname(kobo.__file__), "hub", "static") + '/'



# -*- coding: utf-8 -*-

import six
try:
    import json
except ImportError:
    import simplejson as json

import django.forms.fields
from django.core.exceptions import ValidationError
from django.utils.encoding import force_text
from django.utils.safestring import mark_safe
from django.utils.html import conditional_escape
from django.utils.translation import ugettext_lazy as _
try:
    from django.forms.utils import flatatt
except ImportError:
    from django.forms.util import flatatt


class StateChoiceFormField(django.forms.fields.TypedChoiceField):
    '''
    def __init__(self, *args, **kwargs):
        super(StateChoiceFormField, self).__init__(*args, **kwargs)
         ugly hack - get back reference to StateEnum object
        state = kwargs['coerce'].__self__
        print 'initscff', kwargs['initial'], state, state.choices
        print type(kwargs['initial'])
    '''

    def clean(self, value):
        value = super(StateChoiceFormField, self).clean(value)
        try:
            value = int(str(value))
        except ValueError:
            raise ValidationError('Wrong state value.')
        self.widget.choices = self.choices
        if value in django.forms.fields.EMPTY_VALUES:
            return None
        for c in self.choices:
            if c[0] == value:
                return c[1]
        raise ValidationError('Selected value is not in valid choices.')


class JSONWidget(django.forms.widgets.Textarea):
    def render(self, name, value, attrs=None):
        if value is None: value = ''
        final_attrs = self.build_attrs(attrs, name=name)

        if not isinstance(value, six.string_types):
            value = json.dumps(value)

        return mark_safe(u'<textarea%s>%s</textarea>' % (flatatt(final_attrs),
                conditional_escape(force_text(value))))


class JSONFormField(django.forms.fields.CharField):
    widget = JSONWidget

    def from_db_value(self, value, expression, connection, context):
        return self.to_python(value)


    def to_python(self, value):
        try:
            result = json.loads(value)
        except ValueError:
            raise ValidationError(_("Cannot deserialize JSON data."))
        else:
            if not isinstance(result, (dict, list)):
                raise ValidationError(_("Data types are restricted to JSON serialized dict or list only."))
            return result



from django.contrib.auth.middleware import RemoteUserMiddleware

class LimitedRemoteUserMiddleware(RemoteUserMiddleware):
    '''
    Same behaviour as RemoteUserMiddleware except that it doesn't logout user
    if is already logged in.
    Useful when you have just one authentication powered login page.
    '''
    def process_request(self, request):
        if not hasattr(request, 'user') or not request.user.is_authenticated():
            super(LimitedRemoteUserMiddleware, self).process_request(request)



# -*- coding: utf-8 -*-


class ImproperlyConfigured(Exception):
    """Program is improperly configured."""
    pass


class ShutdownException(Exception):
    """Shutdown currently running program."""
    pass


class AuthenticationError(Exception):
    """Authentication failed for some reason."""
    pass



#!/usr/bin/python
# -*- coding: utf-8 -*-


import re
import six
import unittest2 as unittest

from kobo.tback import get_traceback, Traceback


class TestTraceback(unittest.TestCase):

    def test_empty(self):
        self.assertEqual('', get_traceback())
        self.assertEqual('', Traceback().get_traceback())
        self.assertEqual((None, None, None), Traceback().exc_info)

    def test_text(self):
        try:
            raise Exception('Simple text')
        except:
            str_regexp = re.compile('Traceback \(most recent call last\):\n *File ".*test_tback.py", line .+, in test_text\n *raise Exception\(\'Simple text\'\)\n *Exception: Simple text', re.M)
            bytes_regexp = re.compile(b'Traceback \(most recent call last\):\n *File ".*test_tback.py", line .+, in test_text\n *raise Exception\(\'Simple text\'\)\n *Exception: Simple text', re.M)

            self.assertRegex(get_traceback(), str_regexp)
            tb = Traceback(show_traceback = True, show_code = False, show_locals = False, show_environ = False, show_modules = False)
            self.assertRegex(tb.get_traceback(), bytes_regexp)

    def test_Traceback(self):
        try:
            raise Exception('Simple text')
        except:
            tb = Traceback(show_traceback = False, show_code = False, show_locals = False, show_environ = False, show_modules = False)
        self.assertEqual(b'', tb.get_traceback())
        tb.show_code = True
        self.assertRegex(tb.get_traceback(), re.compile(b'<CODE>.*--> *\d+ *raise Exception.*<\/CODE>$', re.M | re.S))
        tb.show_code = False
        tb.show_locals = True
        self.assertRegex(tb.get_traceback(), re.compile(b'<LOCALS>.*tb = .*<\/LOCALS>$', re.M | re.S))
        tb.show_locals = False
        tb.show_environ = True
        self.assertRegex(tb.get_traceback(), re.compile(b'<ENVIRON>.*<\/ENVIRON>\n<GLOBALS>.*</GLOBALS>$', re.M | re.S))
        tb.show_environ = False
        tb.show_modules = True
        self.assertRegex(tb.get_traceback(), re.compile(b'<MODULES>.*<\/MODULES>$', re.M | re.S))

    def test_encoding(self):
        try:
            a = ''.join([chr(i) for i in range(256)])
            b = u''.join([unichr(i) for i in range(65536)])
            raise Exception()
        except:
            tb = Traceback(show_code = False, show_traceback = False)
        output = tb.get_traceback()
        self.assertIsInstance(output, six.binary_type)

    def test_uninitialized_variables(self):
        class Foo(object):
            __slots__ = ( "bar", "version" )

            def __init__(self):
                self.version = 1

            def test(self):
                try:
                    raise
                except:
                    # bar is uninitialized
                    return Traceback().get_traceback()

        obj = Foo()
        self.assertTrue(obj.test())



# -*- coding: utf-8 -*-


import os

import kobo.cli
import kobo.admin


class Start_Client_Command(kobo.cli.Command):
    """create a hub client command module in the current directory"""
    enabled = True

    def options(self):
        self.parser.usage = "%%prog %s [options] <command-name>" % self.normalized_name
        self.parser.add_option("-d", "--dir", help="target directory")

    def run(self, *args, **kwargs):
        if len(args) < 1:
            self.parser.error("Please specify a name of the command.")

        name = args[0]
        directory = kwargs.pop("dir")
        if not directory:
            directory = os.getcwd()

        try:
            kobo.admin.copy_helper(name, directory, "client@cmd___project_name__.py.template")
        except kobo.admin.TemplateError as ex:
            self.parser.error(ex)



# -*- coding: utf-8 -*-

import base64
import hashlib
import tempfile
import unittest2 as unittest

import django
import pytest
import six

# Only for Django >= 1.7
if 'setup' in dir(django):
    # This has to happen before below imports because they have a hard requirement
    # on settings being loaded before import.
    django.setup()

from datetime import datetime, timedelta

from django.core.exceptions import PermissionDenied
from django.contrib.auth.models import User

from mock import Mock, PropertyMock, patch

from kobo.client.constants import TASK_STATES, FINISHED_STATES
from kobo.exceptions import ShutdownException
from kobo.hub.models import Arch, Channel, Task, Worker
from kobo.hub.xmlrpc import worker

from .utils import DjangoRunner

runner = DjangoRunner()
setup_module = runner.start
teardown_module = runner.stop


def _make_request(w, is_authenticated=True):
    return PropertyMock(
        worker=w,
        user=Mock(is_authenticated=Mock(return_value=is_authenticated)),
    )


class TestXmlRpcWorker(django.test.TransactionTestCase):

    def setUp(self):
        self._fixture_teardown()
        super(TestXmlRpcWorker, self).setUp()

        user = User.objects.create(username='testuser')
        arch = Arch.objects.create(name='testarch')
        channel = Channel.objects.create(name='testchannel')

        w = Worker.objects.create(
            worker_key='testworker',
            name='testworker',
        )

        w.arches.add(arch)
        w.channels.add(channel)
        w.save()

        self._arch = arch
        self._channel = channel
        self._user = user
        self._worker = w

    def test_get_worker_info(self):
        req = _make_request(self._worker)
        wi = worker.get_worker_info(req)

        self.assertFalse(wi is None)
        self.assertEqual(wi['task_count'], 0)
        self.assertEqual(wi['current_load'], 0)
        self.assertEqual(wi['max_load'], 1)
        self.assertTrue(wi['ready'])
        self.assertTrue(wi['enabled'])

    def test_get_worker_id(self):
        req = _make_request(self._worker)
        w_id = worker.get_worker_id(req)

        assert isinstance(w_id, int)

    def test_get_worker_tasks(self):
        for state in TASK_STATES:
            Task.objects.create(
                worker=self._worker,
                arch=self._arch,
                channel=self._channel,
                owner=self._user,
                state=TASK_STATES[state],
            )

        req = _make_request(self._worker)
        tasks = worker.get_worker_tasks(req)

        self.assertEqual(len(tasks), 2)
        self.assertTrue(tasks[0]['id'] < tasks[1]['id'])

        for task in tasks:
            self.assertTrue(task['state'] in [TASK_STATES['ASSIGNED'], TASK_STATES['OPEN']])

    def test_get_worker_tasks_check_wait(self):
        t_parent = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
            waiting=True,
        )

        Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['CLOSED'],
            parent=t_parent,
            awaited=True,
        )

        req = _make_request(self._worker)
        tasks = worker.get_worker_tasks(req)

        self.assertEqual(len(tasks), 1)
        self.assertEqual(tasks[0]['id'], t_parent.id)
        self.assertTrue(tasks[0]['alert'])

    def test_get_worker_tasks_returns_empty_list_if_no_tasks(self):
        req = _make_request(self._worker)
        tasks = worker.get_worker_tasks(req)

        self.assertEqual(tasks, [])

    def test_get_task(self):
        t = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['FREE'],
        )

        req = _make_request(self._worker)
        task_info = worker.get_task(req, t.id)

        self.assertEqual(task_info['id'], t.id)
        self.assertEqual(task_info['worker'], self._worker.id)
        self.assertEqual(task_info['state'], t.state)

    def test_get_task_cant_get_task_from_other_worker(self):
        w = Worker.objects.create(
            worker_key='other-worker',
            name='other-worker',
        )

        t = Task.objects.create(
            worker=w,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['FREE'],
        )

        req = _make_request(self._worker)

        with self.assertRaises(Task.DoesNotExist):
            worker.get_task(req, t.id)

    def test_get_task_no_verify(self):
        t = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['FREE'],
        )

        req = _make_request(self._worker)
        task_info = worker.get_task_no_verify(req, t.id)

        self.assertEqual(task_info['id'], t.id)
        self.assertEqual(task_info['worker'], self._worker.id)
        self.assertEqual(task_info['state'], t.state)

    def test_get_task_no_verify_can_get_task_from_other_worker(self):
        w = Worker.objects.create(
            worker_key='other-worker',
            name='other-worker',
        )

        t = Task.objects.create(
            worker=w,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['FREE'],
        )

        req = _make_request(self._worker)
        task_info = worker.get_task_no_verify(req, t.id)

        self.assertEqual(task_info['id'], t.id)
        self.assertEqual(task_info['worker'], w.id)
        self.assertEqual(task_info['state'], t.state)

    def test_interrupt_tasks_empty_task_list(self):
        req = _make_request(self._worker)
        ok = worker.interrupt_tasks(req, [])
        self.assertTrue(ok)

    def test_interrupt_tasks_single_task(self):
        t = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
        )

        req = _make_request(self._worker)
        ok = worker.interrupt_tasks(req, [t.id])
        self.assertTrue(ok)

        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['INTERRUPTED'])

    def test_interrupt_tasks_multiple_tasks(self):
        t1 = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
        )

        t2 = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
        )

        req = _make_request(self._worker)
        ok = worker.interrupt_tasks(req, [t1.id, t2.id])
        self.assertTrue(ok)

        t1 = Task.objects.get(id=t1.id)
        self.assertEqual(t1.state, TASK_STATES['INTERRUPTED'])

        t2 = Task.objects.get(id=t2.id)
        self.assertEqual(t2.state, TASK_STATES['INTERRUPTED'])

    def test_interrupt_tasks_interrupt_tasks_recursively(self):
        t1 = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
        )

        t2 = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
            parent=t1,
        )

        req = _make_request(self._worker)
        ok = worker.interrupt_tasks(req, [t1.id])
        self.assertTrue(ok)

        t1 = Task.objects.get(id=t1.id)
        self.assertEqual(t1.state, TASK_STATES['INTERRUPTED'])

        t2 = Task.objects.get(id=t2.id)
        self.assertEqual(t2.state, TASK_STATES['INTERRUPTED'])

    def test_interrupt_tasks_fails_to_interrupt_another_worker_task(self):
        w = Worker.objects.create(
            worker_key='other-worker',
            name='other-worker',
        )

        t = Task.objects.create(
            worker=w,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
        )

        req = _make_request(self._worker)

        with self.assertRaises(Task.DoesNotExist):
            worker.interrupt_tasks(req, [t.id])

        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['OPEN'])

    def test_interrupt_tasks_do_not_interrupt_finished_tasks(self):
        tasks = {}

        for state in FINISHED_STATES:
            tasks[state] = Task.objects.create(
                worker=self._worker,
                arch=self._arch,
                channel=self._channel,
                owner=self._user,
                state=state,
            )

        req = _make_request(self._worker)
        ok = worker.interrupt_tasks(req, [t.id for t in tasks.values()])
        self.assertTrue(ok)


        for state, task in tasks.items():
            t = Task.objects.get(id=task.id)
            self.assertEqual(t.state, state)

    @pytest.mark.xfail(six.PY3, reason='Check issue #73 for more info (https://git.io/fxdfm).')
    def test_interrupt_tasks_fails_to_interrupt_if_not_open_or_finished(self):
        tasks = {}

        for state_id in TASK_STATES:
            state = TASK_STATES[state_id]
            if state == TASK_STATES['OPEN'] or state in FINISHED_STATES:
                continue

            tasks[state] = Task.objects.create(
                worker=self._worker,
                arch=self._arch,
                channel=self._channel,
                owner=self._user,
                state=state,
            )

        req = _make_request(self._worker)

        for _, task in tasks.items():
            with self.assertRaises(Exception):
                worker.interrupt_tasks(req, [task.id])

        for state, task in tasks.items():
            t = Task.objects.get(id=task.id)
            self.assertEqual(t.state, state)

    def test_timeout_tasks_empty_task_list(self):
        req = _make_request(self._worker)
        ok = worker.timeout_tasks(req, [])
        self.assertTrue(ok)

    def test_timeout_tasks_single_task(self):
        t = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
        )

        req = _make_request(self._worker)
        ok = worker.timeout_tasks(req, [t.id])
        self.assertTrue(ok)

        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['TIMEOUT'])

    def test_timeout_tasks_multiple_tasks(self):
        t1 = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
        )

        t2 = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
        )

        req = _make_request(self._worker)
        ok = worker.timeout_tasks(req, [t1.id, t2.id])
        self.assertTrue(ok)

        t1 = Task.objects.get(id=t1.id)
        self.assertEqual(t1.state, TASK_STATES['TIMEOUT'])

        t2 = Task.objects.get(id=t2.id)
        self.assertEqual(t2.state, TASK_STATES['TIMEOUT'])

    def test_timeout_tasks_timeout_tasks_recursively(self):
        t1 = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
        )

        t2 = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
            parent=t1,
        )

        req = _make_request(self._worker)
        ok = worker.timeout_tasks(req, [t1.id])
        self.assertTrue(ok)

        t1 = Task.objects.get(id=t1.id)
        self.assertEqual(t1.state, TASK_STATES['TIMEOUT'])

        t2 = Task.objects.get(id=t2.id)
        self.assertEqual(t2.state, TASK_STATES['TIMEOUT'])

    def test_timeout_tasks_fails_to_interrupt_another_worker_task(self):
        w = Worker.objects.create(
            worker_key='other-worker',
            name='other-worker',
        )

        t = Task.objects.create(
            worker=w,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
        )

        req = _make_request(self._worker)

        with self.assertRaises(Task.DoesNotExist):
            worker.timeout_tasks(req, [t.id])

        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['OPEN'])

    def test_timeout_tasks_do_not_interrupt_finished_tasks(self):
        tasks = {}

        for state in FINISHED_STATES:
            tasks[state] = Task.objects.create(
                worker=self._worker,
                arch=self._arch,
                channel=self._channel,
                owner=self._user,
                state=state,
            )

        req = _make_request(self._worker)
        ok = worker.timeout_tasks(req, [t.id for t in tasks.values()])
        self.assertTrue(ok)


        for state, task in tasks.items():
            t = Task.objects.get(id=task.id)
            self.assertEqual(t.state, state)

    @pytest.mark.xfail(six.PY3, reason='Check issue #73 for more info (https://git.io/fxdfm).')
    def test_timeout_tasks_fails_to_interrupt_if_not_open_or_finished(self):
        tasks = {}

        for state_id in TASK_STATES:
            state = TASK_STATES[state_id]
            if state == TASK_STATES['OPEN'] or state in FINISHED_STATES:
                continue

            tasks[state] = Task.objects.create(
                worker=self._worker,
                arch=self._arch,
                channel=self._channel,
                owner=self._user,
                state=state,
            )

        req = _make_request(self._worker)

        for _, task in tasks.items():
            with self.assertRaises(Exception):
                worker.timeout_tasks(req, [task.id])

        for state, task in tasks.items():
            t = Task.objects.get(id=task.id)
            self.assertEqual(t.state, state)

    def test_assign_task(self):
        t = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['FREE'],
        )

        req = _make_request(self._worker)
        worker.assign_task(req, t.id)

        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['ASSIGNED'])

    @pytest.mark.xfail(six.PY3, reason='Check issue #73 for more info (https://git.io/fxdfm).')
    def test_assign_task_fails_to_assing_another_worker_task(self):
        w = Worker.objects.create(
            worker_key='other-worker',
            name='other-worker',
        )

        t = Task.objects.create(
            worker=w,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['FREE'],
        )

        req = _make_request(self._worker)

        with self.assertRaises(Exception):
            worker.assign_task(req, t.id)

        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['FREE'])

    def test_open_task(self):
        t = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['ASSIGNED'],
        )

        req = _make_request(self._worker)
        worker.open_task(req, t.id)

        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['OPEN'])

    @pytest.mark.xfail(six.PY3, reason='Check issue #73 for more info (https://git.io/fxdfm).')
    def test_open_task_fails_to_open_another_worker_task(self):
        w = Worker.objects.create(
            worker_key='other-worker',
            name='other-worker',
        )

        t = Task.objects.create(
            worker=w,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['ASSIGNED'],
        )

        req = _make_request(self._worker)

        with self.assertRaises(Exception):
            worker.open_task(req, t.id)

        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['ASSIGNED'])

    def test_close_task(self):
        t = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
        )

        req = _make_request(self._worker)
        worker.close_task(req, t.id, '')

        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['CLOSED'])

    def test_close_task_fails_to_close_another_worker_task(self):
        w = Worker.objects.create(
            worker_key='other-worker',
            name='other-worker',
        )

        t = Task.objects.create(
            worker=w,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
        )

        req = _make_request(self._worker)

        with self.assertRaises(Task.DoesNotExist):
            worker.close_task(req, t.id, '')

        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['OPEN'])

    def test_cancel_task(self):
        t = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
        )

        req = _make_request(self._worker)
        worker.cancel_task(req, t.id)

        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['CANCELED'])

    def test_cancel_task_fails_to_cancel_another_worker_task(self):
        w = Worker.objects.create(
            worker_key='other-worker',
            name='other-worker',
        )

        t = Task.objects.create(
            worker=w,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
        )

        req = _make_request(self._worker)

        with self.assertRaises(Task.DoesNotExist):
            worker.cancel_task(req, t.id)

        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['OPEN'])

    def test_fail_task(self):
        t = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
        )

        req = _make_request(self._worker)
        worker.fail_task(req, t.id, '')

        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['FAILED'])

    def test_fail_task_fails_to_fail_another_worker_task(self):
        w = Worker.objects.create(
            worker_key='other-worker',
            name='other-worker',
        )

        t = Task.objects.create(
            worker=w,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
        )

        req = _make_request(self._worker)

        with self.assertRaises(Task.DoesNotExist):
            worker.fail_task(req, t.id, '')

        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['OPEN'])

    def test_set_task_weight(self):
        t = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
        )

        self.assertEqual(t.weight, 1)

        req = _make_request(self._worker)
        weight = worker.set_task_weight(req, t.id, 100)
        self.assertEqual(weight, 100)

        t = Task.objects.get(id=t.id)
        self.assertEqual(t.weight, 100)

    def test_set_task_weight_fails_if_another_worker_task(self):
        w = Worker.objects.create(
            worker_key='other-worker',
            name='other-worker',
        )

        t = Task.objects.create(
            worker=w,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
        )

        self.assertEqual(t.weight, 1)

        req = _make_request(self._worker)

        with self.assertRaises(Task.DoesNotExist):
            worker.set_task_weight(req, t.id, 100)

        t = Task.objects.get(id=t.id)
        self.assertEqual(t.weight, 1)

    @pytest.mark.xfail(reason='Check issue #68 for more info (https://git.io/fxSZ2).')
    def test_update_worker(self):
        req = _make_request(self._worker)

        self.assertTrue(self._worker.enabled)
        self.assertTrue(self._worker.ready)
        self.assertEqual(self._worker.task_count, 0)

        wi = worker.update_worker(req, False, False, 1)

        self.assertFalse(wi['enabled'])
        self.assertFalse(wi['ready'])
        self.assertEqual(wi['task_count'], 1)

    def test_get_tasks_to_assign(self):
        for _ in range(2):
            Task.objects.create(
                worker=self._worker,
                arch=self._arch,
                channel=self._channel,
                owner=self._user,
                state=TASK_STATES['FREE'],
            )

            Task.objects.create(
                worker=self._worker,
                arch=self._arch,
                channel=self._channel,
                owner=self._user,
                state=TASK_STATES['ASSIGNED'],
                exclusive=True,
            )

            Task.objects.create(
                worker=self._worker,
                arch=self._arch,
                channel=self._channel,
                owner=self._user,
                state=TASK_STATES['FREE'],
                awaited=True,
            )

            Task.objects.create(
                worker=self._worker,
                arch=self._arch,
                channel=self._channel,
                owner=self._user,
                state=TASK_STATES['ASSIGNED'],
            )

            Task.objects.create(
                worker=self._worker,
                arch=self._arch,
                channel=self._channel,
                owner=self._user,
                state=TASK_STATES['CLOSED'],
            )

        req = _make_request(self._worker)
        tasks = worker.get_tasks_to_assign(req)

        self.assertEqual(len(tasks), 8)
        self.assertEqual(len([t for t in tasks if t['state'] == TASK_STATES['FREE'] and not t['exclusive'] and not t['awaited']]), 2)
        self.assertEqual(len([t for t in tasks if t['state'] == TASK_STATES['ASSIGNED'] and t['exclusive']]), 2)
        self.assertEqual(len([t for t in tasks if t['state'] == TASK_STATES['FREE'] and t['awaited']]), 2)
        self.assertEqual(len([t for t in tasks if t['state'] == TASK_STATES['ASSIGNED'] and not t['exclusive']]), 2)

    def test_get_tasks_to_assign_limit_tasks(self):
        for _ in range(10):
            Task.objects.create(
                worker=self._worker,
                arch=self._arch,
                channel=self._channel,
                owner=self._user,
                state=TASK_STATES['FREE'],
            )

            Task.objects.create(
                worker=self._worker,
                arch=self._arch,
                channel=self._channel,
                owner=self._user,
                state=TASK_STATES['ASSIGNED'],
                exclusive=True,
            )

            Task.objects.create(
                worker=self._worker,
                arch=self._arch,
                channel=self._channel,
                owner=self._user,
                state=TASK_STATES['FREE'],
                awaited=True,
            )

        req = _make_request(self._worker)
        tasks = worker.get_tasks_to_assign(req)

        self.assertEqual(len(tasks), 10)
        self.assertEqual(len([t for t in tasks if t['state'] == TASK_STATES['ASSIGNED'] and t['exclusive']]), 10)

    def test_get_awaited_tasks(self):
        t1 = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['FREE'],
        )

        t2 = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['FREE'],
            awaited=True,
            parent=t1,
        )

        req = _make_request(self._worker)
        tasks = worker.get_awaited_tasks(req, [t1.export()])
        self.assertEqual(len(tasks), 1)
        self.assertEqual(tasks[0]['id'], t2.id)

    def test_get_awaited_tasks_if_empty_task_list(self):
        t_parent = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['FREE'],
        )

        Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['FREE'],
            awaited=True,
            parent=t_parent,
        )

        req = _make_request(self._worker)
        tasks = worker.get_awaited_tasks(req, [])
        self.assertEqual(len(tasks), 0)

    def test_create_subtask(self):
        t_parent = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['FREE'],
        )

        req = _make_request(self._worker)
        task_id = worker.create_subtask(req, 'Label', 'Method', None, t_parent.id)
        self.assertTrue(task_id > 0)

        t_child = Task.objects.get(id=task_id)
        self.assertEqual(t_child.parent.id, t_parent.id)
        self.assertEqual(t_child.label, 'Label')
        self.assertEqual(t_child.method, 'Method')

    def test_create_subtask_if_another_worker_task(self):
        w = Worker.objects.create(
            worker_key='other-worker',
            name='other-worker',
        )

        t = Task.objects.create(
            worker=w,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['FREE'],
        )

        req = _make_request(self._worker)

        with self.assertRaises(Task.DoesNotExist):
            worker.create_subtask(req, '', '', None, t.id)

    def test_wait(self):
        t = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['FREE'],
        )

        self.assertFalse(t.waiting)

        req = _make_request(self._worker)
        ok = worker.wait(req, t.id)
        self.assertTrue(ok)

        t = Task.objects.get(id=t.id)
        self.assertTrue(t.waiting)

    def test_wait_subtasks(self):
        t1 = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['FREE'],
        )

        t2 = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['FREE'],
            parent=t1,
        )

        self.assertFalse(t1.waiting)
        self.assertFalse(t1.awaited)
        self.assertFalse(t2.waiting)
        self.assertFalse(t2.awaited)

        req = _make_request(self._worker)
        ok = worker.wait(req, t1.id)
        self.assertTrue(ok)

        t1 = Task.objects.get(id=t1.id)
        self.assertTrue(t1.waiting)
        self.assertFalse(t1.awaited)

        t2 = Task.objects.get(id=t2.id)
        self.assertFalse(t2.waiting)
        self.assertTrue(t2.awaited)

    def test_wait_subtasks_filter_child(self):
        t1 = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['FREE'],
        )

        t2 = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['FREE'],
            parent=t1,
        )

        t3 = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['FREE'],
            parent=t1,
        )

        self.assertFalse(t1.waiting)
        self.assertFalse(t1.awaited)
        self.assertFalse(t2.waiting)
        self.assertFalse(t2.awaited)
        self.assertFalse(t3.waiting)
        self.assertFalse(t3.awaited)

        req = _make_request(self._worker)
        ok = worker.wait(req, t1.id, [t2.id])
        self.assertTrue(ok)

        t1 = Task.objects.get(id=t1.id)
        self.assertTrue(t1.waiting)
        self.assertFalse(t1.awaited)

        t2 = Task.objects.get(id=t2.id)
        self.assertFalse(t2.waiting)
        self.assertTrue(t2.awaited)

        t3 = Task.objects.get(id=t3.id)
        self.assertFalse(t3.waiting)
        self.assertFalse(t3.awaited)

    def test_check_wait(self):
        t_parent = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
        )

        Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['CLOSED'],
            parent=t_parent,
        )

        Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['CLOSED'],
            parent=t_parent,
        )

        req = _make_request(self._worker)
        finished, unfinished = worker.check_wait(req, t_parent.id)
        self.assertEqual(len(finished), 2)
        self.assertEqual(len(unfinished), 0)

    def test_check_wait_subtasks_not_finished(self):
        t_parent = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['FREE'],
        )

        Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['FREE'],
            parent=t_parent,
        )

        Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['FREE'],
            parent=t_parent,
        )

        req = _make_request(self._worker)
        finished, unfinished = worker.check_wait(req, t_parent.id)
        self.assertEqual(len(finished), 0)
        self.assertEqual(len(unfinished), 2)

    def test_check_wait_without_subtasks(self):
        t = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['FREE'],
        )

        req = _make_request(self._worker)
        finished, unfinished = worker.check_wait(req, t.id)
        self.assertEqual(len(finished), 0)
        self.assertEqual(len(unfinished), 0)

    def test_upload_task_log(self):
        t = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
        )

        req = _make_request(self._worker)

        with tempfile.NamedTemporaryFile(delete=True) as tf:
            msg = b'this is a text message'
            checksum = hashlib.sha256(msg).hexdigest()
            chunk = base64.encodestring(msg)
            chunk_start = 0
            chunk_size = len(msg)

            ok = worker.upload_task_log(req,
                                        t.id,
                                        tf.name,
                                        0o644,
                                        str(chunk_start),
                                        str(chunk_size),
                                        checksum, chunk)

            self.assertTrue(ok)
            self.assertEqual(tf.read(), msg)

    def test_upload_task_log_catch_decode_error(self):
        t = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
        )

        req = _make_request(self._worker)

        with tempfile.NamedTemporaryFile(delete=True) as tf:
            msg = b'this is a text message'
            checksum = 'invalid-checksum'
            chunk = base64.encodestring(msg)
            chunk_start = 0
            chunk_size = len(msg)

            ok = worker.upload_task_log(req,
                                        t.id,
                                        tf.name,
                                        0o644,
                                        str(chunk_start),
                                        str(chunk_size),
                                        checksum, chunk)

            self.assertFalse(ok)
            self.assertEqual(tf.read(), b'')

    def test_upload_task_log_fails_if_invalid_path(self):
        t = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['OPEN'],
        )

        req = _make_request(self._worker)

        with self.assertRaises(ValueError):
            worker.upload_task_log(req, t.id, '../foo/bar', 0o644, '', '', '', '')

    def test_upload_task_log_fails_if_task_not_open(self):
        t = Task.objects.create(
            worker=self._worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            state=TASK_STATES['CLOSED'],
        )

        req = _make_request(self._worker)

        with self.assertRaises(ValueError):
            worker.upload_task_log(req, t.id, '/foo/bar', 0o644, '', '', '', '')


class TestXmlRpcWorkerNotAuthenticated(django.test.TransactionTestCase):

    def test_get_worker_info(self):
        with self.assertRaises(PermissionDenied):
            worker.get_worker_info(_make_request(None, False))

    def test_get_worker_id(self):
        with self.assertRaises(PermissionDenied):
            worker.get_worker_id(_make_request(None, False))

    def test_get_worker_tasks(self):
        with self.assertRaises(PermissionDenied):
            worker.get_worker_tasks(_make_request(None, False))

    def test_get_task(self):
        with self.assertRaises(PermissionDenied):
            worker.get_task(_make_request(None, False), 1)

    def test_get_task_no_verify(self):
        with self.assertRaises(PermissionDenied):
            worker.get_task_no_verify(_make_request(None, False), 1)

    def test_interrupt_tasks(self):
        with self.assertRaises(PermissionDenied):
            worker.interrupt_tasks(_make_request(None, False), [])

    def test_timeout_tasks(self):
        with self.assertRaises(PermissionDenied):
            worker.timeout_tasks(_make_request(None, False), [])

    def test_assign_task(self):
        with self.assertRaises(PermissionDenied):
            worker.assign_task(_make_request(None, False), 1)

    def test_open_task(self):
        with self.assertRaises(PermissionDenied):
            worker.open_task(_make_request(None, False), 1)

    def test_close_task(self):
        with self.assertRaises(PermissionDenied):
            worker.close_task(_make_request(None, False), 1, None)

    def test_cancel_task(self):
        with self.assertRaises(PermissionDenied):
            worker.cancel_task(_make_request(None, False), 1)

    def test_fail_task(self):
        with self.assertRaises(PermissionDenied):
            worker.fail_task(_make_request(None, False), 1, None)

    def test_set_task_weight(self):
        with self.assertRaises(PermissionDenied):
            worker.set_task_weight(_make_request(None, False), 1, 10)

    def test_update_worker(self):
        with self.assertRaises(PermissionDenied):
            worker.update_worker(_make_request(None, False), True, True, 0)

    def test_get_tasks_to_assign(self):
        with self.assertRaises(PermissionDenied):
            worker.get_tasks_to_assign(_make_request(None, False))

    def test_get_awaited_tasks(self):
        with self.assertRaises(PermissionDenied):
            worker.get_awaited_tasks(_make_request(None, False), [])

    def test_create_subtask(self):
        with self.assertRaises(PermissionDenied):
            worker.create_subtask(_make_request(None, False), '', '', None, 1)

    def test_wait(self):
        with self.assertRaises(PermissionDenied):
            worker.wait(_make_request(None, False), 1)

    def test_check_wait(self):
        with self.assertRaises(PermissionDenied):
            worker.check_wait(_make_request(None, False), 1)

    def test_upload_task_log(self):
        with self.assertRaises(PermissionDenied):
            worker.upload_task_log(_make_request(None, False), 1, '/dev/null', 0o666, 0, 0, None, None)



# -*- coding: utf-8 -*-


# Some code comes from koji: https://fedorahosted.org/koji/


from six.moves import zip as izip
import datetime
import time
import six
from six.moves import range
try:
    import email.utils as email_utils
except ImportError:
    import email.Utils as email_utils

import koji
import rpm


__all__ = (
    "FILE_DIGEST_ALGO_MAP",
    "compare_nvr",
    "get_rpm_header",
    "get_header_field",
    "get_header_fields",
    "split_nvr_epoch",
    "make_nvr",
    "make_nvra",
    "parse_nvr",
    "parse_nvra",
    "parse_evr",
    "get_keys_from_header",
    "get_digest_algo_from_header",
    "get_changelogs_from_header",
)


body_header_tags = ["siggpg", "sigpgp"]
head_header_tags = ["dsaheader", "rsaheader"]


FILE_DIGEST_ALGO_MAP = {
    1: "MD5",
    2: "SHA1",
    8: "SHA256",
    9: "SHA384",
    10: "SHA512",
}



def get_rpm_header(file_name, ts=None):
    """Read rpm header.

    @param file_name: rpm file name (or file object)
    @type file_name: str (or file)
    @param ts: transaction set instance
    @type ts: rpm.TransactionSet
    @return: rpm header
    @rtype: rpm.hdr
    """

    if ts is None:
        ts = rpm.TransactionSet()
        # no rpm.keyring on rhel5
        if hasattr(rpm, "keyring"):
            # Set an empty keyring to prevent accessing rpmdb,
            # which may cause race-conditions when running in threads.
            # NOTE: RPM is *not* tread-safe, but this *usually* works in threads.
            ts.setKeyring(rpm.keyring())
        ts.setVSFlags(rpm._RPMVSF_NOSIGNATURES|rpm._RPMVSF_NODIGESTS)

    if isinstance(file_name, six.string_types):
        fo = open(file_name, "r")
    else:
        fo = file_name

    hdr = ts.hdrFromFdno(fo.fileno())

    if fo is not file_name:
        fo.close()

    return hdr


def get_header_field(hdr, name, decode=True):
    """Extract named field from a rpm header.

    @param hdr: a rpm header
    @type hdr: rpm.hdr
    @param name: a rpm field name
    @type name: str
    @param decode: whether to decode returned bytes to str
    @type decode: bool
    @return: value of a rpm field
    @rtype: str or int or list
    """

    if name == "arch":
        # HACK: return "src" or "nosrc" arch instead of build arch
        if get_header_field(hdr, "sourcepackage"):
            if get_header_field(hdr, "nosource") or get_header_field(hdr, "nopatch"):
                return "nosrc"
            return "src"

    hdr_key = getattr(rpm, "RPMTAG_%s" % name.upper(), None)
    if hdr_key is None:
        # HACK: nosource and nopatch may not be in exported rpm tags
        if name == "nosource":
            hdr_key = 1051
        elif name == "nopatch":
            hdr_key = 1052
        else:
            raise AttributeError("No such rpm header field: %s" % name)

    result = hdr[hdr_key]
    if name in ("nosource", "nopatch"):
        # HACK: workaround for https://bugzilla.redhat.com/show_bug.cgi?id=991329
        if result is None:
            result = []
        elif isinstance(result, six.integer_types):
            result = [result]

    if six.PY3 and decode:
        # We are on Python 3 with python3-rpm <= 4.14.2 that returns bytes, the
        # value needs to be decoded. Newer versions of python3-rpm will do the
        # decoding by themselves and don't need to be handled here.
        if isinstance(result, bytes):
            result = result.decode("utf-8", "surrogateescape")
        if isinstance(result, list):
            result = [
                x.decode("utf-8", "surrogateescape") if isinstance(x, bytes) else x
                for x in result
            ]
    return result


def get_header_fields(hdr, fields):
    """Extract named fields from a rpm header and return as a dictionary.

    @param fields: rpm field names
    @type fields: list
    @rtype: dict
    """

    result = {}
    for field in fields:
        result[field] = get_header_field(hdr, field)
    return result


def split_nvr_epoch(nvre):
    """Split nvre to N-V-R and E.

    @param nvre: E:N-V-R or N-V-R:E string
    @type nvre: str
    @return: (N-V-R, E)
    @rtype: (str, str)
    """

    if ":" in nvre:
        if nvre.count(":") != 1:
            raise ValueError("Invalid NVRE: %s" % nvre)

        nvr, epoch = nvre.rsplit(":", 1)
        if "-" in epoch:
            if "-" not in nvr:
                # switch nvr with epoch
                nvr, epoch = epoch, nvr
            else:
                # it's probably N-E:V-R format, handle it after the split
                nvr, epoch = nvre, ""
    else:
        nvr, epoch = nvre, ""

    return (nvr, epoch)


def parse_nvr(nvre):
    """Split N-V-R into a dictionary.

    @param nvre: N-V-R:E, E:N-V-R or N-E:V-R string
    @type nvre: str
    @return: {name, version, release, epoch}
    @rtype: dict
    """

    if "/" in nvre:
        nvre = nvre.split("/")[-1]

    nvr, epoch = split_nvr_epoch(nvre)

    nvr_parts = nvr.rsplit("-", 2)
    if len(nvr_parts) != 3:
        raise ValueError("Invalid NVR: %s" % nvr)

    # parse E:V
    if epoch == "" and ":" in nvr_parts[1]:
        epoch, nvr_parts[1] = nvr_parts[1].split(":", 1)

    # check if epoch is empty or numeric
    if epoch != "":
        try:
            int(epoch)
        except ValueError:
            raise ValueError("Invalid epoch '%s' in '%s'" % (epoch, nvr))

    result = dict(zip(["name", "version", "release"], nvr_parts))
    result["epoch"] = epoch
    return result


def parse_nvra(nvra):
    """Split N-V-R.A[.rpm] into a dictionary.

    @param nvra: N-V-R:E.A[.rpm], E:N-V-R.A[.rpm], N-V-R.A[.rpm]:E or N-E:V-R.A[.rpm] string
    @type nvra: str
    @return: {name, version, release, epoch, arch}
    @rtype: dict
    """

    if "/" in nvra:
        nvra = nvra.split("/")[-1]

    epoch = ""
    for i in range(2):
        # run this twice to parse N-V-R.A.rpm:E and N-V-R.A:E.rpm
        if nvra.endswith(".rpm"):
            # strip .rpm suffix
            nvra = nvra[:-4]
        else:
            # split epoch (if exists)
            nvra, epoch = split_nvr_epoch(nvra)

    nvra_parts = nvra.rsplit(".", 1)
    if len(nvra_parts) != 2:
        raise ValueError("Invalid NVRA: %s" % nvra)

    nvr, arch = nvra_parts
    if "-" in arch:
        raise ValueError("Invalid arch '%s' in '%s'" % (arch, nvra))

    result = parse_nvr(nvr)
    result["arch"] = arch
    result["src"] = (arch == "src")
    if epoch != "":
        result["epoch"] = epoch
    return result


def parse_evr(evr, allow_empty_release=False):
    """Split E:V-R into a dictionary."""

    if ":" in evr:
        epoch, vr = evr.split(":", 1)
        try:
            int(epoch)
        except ValueError:
            try:
                int(vr)
            except ValueError:
                raise ValueError("Invalid epoch in '%s'" % evr)
            epoch, vr = vr, epoch
    else:
        epoch, vr = ("", evr)

    if "-" in vr:
        version, release = vr.split("-", 1)
    else:
        if not allow_empty_release:
            raise ValueError("Missing release in '%s'" % evr)
        version, release = (vr, "")

    result = {
        "epoch": epoch,
        "version": version,
        "release": release,
    }

    return result


def make_nvr(nvrea_dict, add_epoch=False, force_epoch=False):
    """Make [E:]N-V-R from a nvrea dictionary.

    @param nvrea_dict: {name, version, release, epoch}
    @type nvrea_dict: dict
    @param add_epoch: add epoch to the result
    @type add_epoch: bool
    @return: N-[E:]V-R string
    @rtype str
    """

    if add_epoch or force_epoch:
        epoch = nvrea_dict.get("epoch", None)
        epoch = epoch or 0
        epoch = int(epoch)
        if epoch or force_epoch:
            epoch = "%s:" % epoch
        else:
            epoch = ""
    else:
        epoch = ""

    return "%s-%s%s-%s" % (nvrea_dict["name"], epoch, nvrea_dict["version"], nvrea_dict["release"])


def make_nvra(nvrea_dict, add_epoch=False, force_epoch=False, add_rpm=False):
    """Make N-[E:]V-R.A[.rpm] from a nvrea dictionary.

    @param nvrea_dict: {name, version, release, epoch, arch}
    @type nvrea_dict: dict
    @param add_epoch: add epoch to the result
    @type add_epoch: bool
    @param add_rpm: append '.rpm' suffix
    @type add_rpm: bool
    @return: [E:]N-V-R.A[.rpm] string
    @rtype str
    """

    result = "%s.%s" % (make_nvr(nvrea_dict, add_epoch, force_epoch), nvrea_dict["arch"])
    if add_rpm:
        result += ".rpm"
    return result


def make_nvrea_list(nvrea_dict):
    """Make [N, V, R, E, A] list from a nvrea dictionary.

    @param nvrea_dict: {name, version, release, epoch, arch}
    @type nvrea_dict: dict
    @return: [name, version, release, epoch, arch]
    @rtype: str
    """
    return [ nvrea_dict[i] for i in ("name", "version", "release", "epoch", "arch") ]


def compare_nvr(nvr_dict1, nvr_dict2, ignore_epoch=False):
    """Compare two N-V-R dictionaries.

    @param nvr_dict1: {name, version, release, epoch}
    @type nvr_dict1: dict
    @param nvr_dict2: {name, version, release, epoch}
    @type nvr_dict2: dict
    @param ignore_epoch: ignore epoch during the comparison
    @type ignore_epoch: bool
    @return: nvr1 newer than nvr2: 1, same nvrs: 0, nvr1 older: -1, different names: ValueError
    @rtype: int
    """

    nvr1 = nvr_dict1.copy()
    nvr2 = nvr_dict2.copy()

    nvr1["epoch"] = nvr1.get("epoch", None)
    nvr2["epoch"] = nvr2.get("epoch", None)

    if nvr1["name"] != nvr2["name"]:
        raise ValueError("Package names doesn't match: %s, %s" % (nvr1["name"], nvr2["name"]))

    if ignore_epoch:
        nvr1["epoch"] = 0
        nvr2["epoch"] = 0

    if nvr1["epoch"] is None:
        nvr1["epoch"] = ""

    if nvr2["epoch"] is None:
        nvr2["epoch"] = ""

    return rpm.labelCompare((str(nvr1["epoch"]), str(nvr1["version"]), str(nvr1["release"])), (str(nvr2["epoch"]), str(nvr2["version"]), str(nvr2["release"])))

def get_keys_from_header(hdr):
    """Extract signing key id from a rpm header.

    @param hdr: rpm header
    @type hdr: rpm.hdr
    @return: signing key id represented as an uppercase hex string
    @rtype: str
    """

    result = []
    head_keys = []

    for field in head_header_tags:
        sigkey = get_header_field(hdr, field, decode=False)
        if sigkey:
            head_keys.append(koji.get_sigpacket_key_id(sigkey).upper())

    for field in body_header_tags:
        sigkey = get_header_field(hdr, field, decode=False)
        if sigkey:
            key_id = koji.get_sigpacket_key_id(sigkey).upper()
            if key_id in head_keys:
                result.append(key_id)
            else:
                raise ValueError("%s key not found in head keys: %s" % (field, key_id))

    if len(result) > 1:
        raise ValueError("More than one key found: %s" % result)

    if len(result) == 1:
        return result[0]


def get_digest_algo_from_header(hdr):
    """Read file digest algorithm from a rpm header.

    @param hdr: rpm header
    @type hdr: rpm.hdr
    @return: digest algorithm name
    @rtype: str
    """

    # RPMTAG_FILEDIGESTALGO is not defined in older rpm
    hdr_key = getattr(rpm, "RPMTAG_FILEDIGESTALGO", 5011)
    algo_id = hdr[hdr_key]

    if algo_id == [] or algo_id is None:
        # RPMTAG_FILEDIGESTALGO is empty, fall back to md5
        algo_id = 1

    if algo_id not in FILE_DIGEST_ALGO_MAP:
        raise ValueError("Unknown file digest algorithm id: %s" % algo_id)

    return FILE_DIGEST_ALGO_MAP[algo_id]


def get_file_list_from_header(hdr):
    """Read file list from a rpm header.

    @param hdr: rpm header
    @type hdr: rpm.hdr
    @return: file_name: (color, checksum)
    @rtype: dict
    """

    result = {}
    fi = hdr.fiFromHeader()
    for file_obj in fi:
        name = file_obj[0]
        result[name] = (fi.FColor(), fi.MD5())
    return result


class ChangelogEntry(object):
    def __init__(self, ch_name, ch_time, ch_text):
        self.name = ch_name
        self.time = ch_time
        self.text = ch_text

    def __str__(self):
        return "* %s %s\n%s" % (self.ctime, self.name, self.text)

    def format_time(self, fmt):
        return time.strftime(fmt, time.gmtime(self.time))

    @property
    def ctime(self):
        return self.format_time("%a %b %d %Y")

    @property
    def iso_date(self):
        return self.format_time("%F")

    @property
    def author(self):
        return email_utils.parseaddr(self.name)[0]

    @property
    def email(self):
        return email_utils.parseaddr(self.name)[1]


def get_changelogs_from_header(hdr, max_records=None, newer_than=None):
    """Read changelogs from a rpm header.

    @param hdr: rpm header
    @type hdr: rpm.hdr
    @param max_records: limit output to max_records
    @type max_records: int
    @param newer_than: return records newer than YYYY-MM-DD
    @type newer_than: str
    @return: [ChangelogEntry()]
    @rtype: list
    """

    if max_records < 0:
        max_records = None
    if max_records == 0:
        return []

    names = get_header_field(hdr, "CHANGELOGNAME")
    times = get_header_field(hdr, "CHANGELOGTIME")
    texts = get_header_field(hdr, "CHANGELOGTEXT")

    if isinstance(newer_than, (float, int)):
        newer_than = int(newer_than)
    elif isinstance(newer_than, str):
        newer_than = int(datetime.datetime(*(time.strptime(newer_than, "%Y-%m-%d")[0:6])).strftime("%s"))
    elif isinstance(newer_than, datetime.datetime):
        newer_than = int(newer_than.strftime("%s"))

    result = []
    for num, (ch_name, ch_time, ch_text) in enumerate(izip(names, times, texts)):
        if max_records is not None and num >= max_records:
            break
        if newer_than is not None and ch_time < newer_than:
            break
        log = ChangelogEntry(ch_name, ch_time, ch_text)
        result.append(log)
    return result
